@@ -52,6 +52,8 @@ OR PERFORMANCE OF THIS SOFTWARE.
 
 ******************************************************************/
 
+#include "config.h"
+
 #include <errno.h>
 
 #include "utils.h"
@@ -144,7 +146,7 @@ skip_more_whitespace_and_comments:
 
     /* LHS Keysym. */
     if (chr(s, '<')) {
-        while (peek(s) != '>' && !eol(s))
+        while (peek(s) != '>' && !eol(s) && !eof(s))
             buf_append(s, next(s));
         if (!chr(s, '>')) {
             scanner_err(s, "unterminated keysym literal");
@@ -316,6 +318,7 @@ struct production {
     unsigned int len;
     xkb_keysym_t keysym;
     char string[256];
+    /* At least one of these is true. */
     bool has_keysym;
     bool has_string;
 
@@ -400,9 +403,20 @@ add_production(struct xkb_compose_table *table, struct scanner *s,
     }
 
     if (node->u.leaf.utf8 != 0 || node->u.leaf.keysym != XKB_KEY_NoSymbol) {
-        if (streq(&darray_item(table->utf8, node->u.leaf.utf8),
-                  production->string) &&
-            node->u.leaf.keysym == production->keysym) {
+        bool same_string =
+            (node->u.leaf.utf8 == 0 && !production->has_string) ||
+            (
+                node->u.leaf.utf8 != 0 && production->has_string &&
+                streq(&darray_item(table->utf8, node->u.leaf.utf8),
+                      production->string)
+            );
+        bool same_keysym =
+            (node->u.leaf.keysym == XKB_KEY_NoSymbol && !production->has_keysym) ||
+            (
+                node->u.leaf.keysym != XKB_KEY_NoSymbol && production->has_keysym &&
+                node->u.leaf.keysym == production->keysym
+            );
+        if (same_string && same_keysym) {
             scanner_warn(s, "this compose sequence is a duplicate of another; skipping line");
             return;
         }
@@ -464,7 +478,7 @@ do_include(struct xkb_compose_table *table, struct scanner *s,
         return false;
     }
 
-    file = fopen(path, "r");
+    file = fopen(path, "rb");
     if (!file) {
         scanner_err(s, "failed to open included Compose file \"%s\": %s",
                     path, strerror(errno));
@@ -655,7 +669,7 @@ rhs:
         }
         production.keysym = keysym;
         production.has_keysym = true;
-	/* fallthrough */
+        /* fallthrough */
     case TOK_END_OF_LINE:
         if (!production.has_string && !production.has_keysym) {
             scanner_warn(s, "right-hand side must have at least one of string or keysym; skipping line");
