@@ -32,17 +32,18 @@
 
 #include "darray.h"
 
-/*
- * We sometimes malloc strings and then expose them as const char*'s. This
- * macro is used when we free these strings in order to avoid -Wcast-qual
- * errors.
- */
-#define UNCONSTIFY(const_ptr)  ((void *) (uintptr_t) (const_ptr))
-
 #define STATIC_ASSERT(expr, message) do { \
     switch (0) { case 0: case (expr): ; } \
 } while (0)
 
+#define ARRAY_SIZE(arr) ((sizeof(arr) / sizeof(*(arr))))
+
+#define MIN(a, b) ((a) < (b) ? (a) : (b))
+#define MAX(a, b) ((a) > (b) ? (a) : (b))
+
+/* Round up @a so it's divisible by @b. */
+#define ROUNDUP(a, b) (((a) + (b) - 1) / (b) * (b))
+
 char
 to_lower(char c);
 
@@ -117,17 +118,20 @@ memdup(const void *mem, size_t nmemb, size_t size)
     return p;
 }
 
-static inline int
-min(int misc, int other)
+#if !(defined(HAVE_STRNDUP) && HAVE_STRNDUP)
+static inline char *
+strndup(const char *s, size_t n)
 {
-    return (misc < other) ? misc : other;
-}
-
-static inline int
-max(int misc, int other)
-{
-    return (misc > other) ? misc : other;
+    size_t slen = strlen(s);
+    size_t len = MIN(slen, n);
+    char *p = malloc(len + 1);
+    if (!p)
+        return NULL;
+    memcpy(p, s, len);
+    p[len] = '\0';
+    return p;
 }
+#endif
 
 /* ctype.h is locale-dependent and has other oddities. */
 static inline bool
@@ -198,16 +202,6 @@ map_file(FILE *file, char **string_out, size_t *size_out);
 void
 unmap_file(char *string, size_t size);
 
-#define ARRAY_SIZE(arr) ((sizeof(arr) / sizeof(*(arr))))
-
-#define MIN(a, b) ((a) < (b) ? (a) : (b))
-#define MIN3(a, b, c) MIN(MIN((a), (b)), (c))
-#define MAX(a, b) ((a) > (b) ? (a) : (b))
-#define MAX3(a, b, c) MAX(MAX((a), (b)), (c))
-
-/* Round up @a so it's divisible by @b. */
-#define ROUNDUP(a, b) (((a) + (b) - 1) / (b) * (b))
-
 #if defined(HAVE_SECURE_GETENV)
 # define secure_getenv secure_getenv
 #elif defined(HAVE___SECURE_GETENV)
@@ -265,4 +259,12 @@ unmap_file(char *string, size_t size);
 #define ATTR_PACKED
 #endif
 
+#if !(defined(HAVE_ASPRINTF) && HAVE_ASPRINTF)
+int asprintf(char **strp, const char *fmt, ...) ATTR_PRINTF(2, 3);
+# if !(defined(HAVE_VASPRINTF) && HAVE_VASPRINTF)
+#  include <stdarg.h>
+int vasprintf(char **strp, const char *fmt, va_list ap);
+# endif /* !HAVE_VASPRINTF */
+#endif /* !HAVE_ASPRINTF */
+
 #endif /* UTILS_H */
